/*
 * Copyright (C) 1996 by Uwe Bonnes
 *
 * This code read /proc/smcio and give some values in a human readable form
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>


#define NO_SMC_REGS 16


static char *version = "Scan-Smcio Version 0.1 961201";


void print_usage(char *progname) {
  printf("Usage: %s\n\n", progname);
  printf("Change values with\n");
  printf("\techo \"smcio <keyword> <values> > /proc/smcio\"\n");
  printf("\tpossible keywords and values are (* denotes reset defaults)\n");
  printf("\tppadr 0x[0|3BC|378|278(*)]");
  printf("\tppmode [spp(*)|epp|ecp|all]\n");
  printf("\tppepptype 1.[7|9(*)]\n");
  printf("\tppfifothr (val) (*=0)\n");
  printf("\tuart1addr 0x[3F8(*)|2F8|338|238|3E8|2E8|2E0|220|228]\n");
  printf("\tuart1en enable(*)|disable\n");
  printf("\tuart1pwr on(*)|off\n");
  printf("\tuart2addr 0x[3F8|2F8(*)|338|238|3E8|2E8|2E0|220|228]\n");
  printf("\tuart2en enable(*)|disable\n");
  printf("\tuart2pwr on(*)|off\n");
  printf("\t\n");
  printf("\tFor the serial port, only some combinations for uart1 and 2 are allowed\n");
  printf("\tSetting invalid combinations is blocked\n");
  printf("\tFor deadlocks when changing ports, disable one port\n");
  printf("\tEnable a tty for the non standard ports with:\n");
  printf("\t\"setserial /dev/ttyS1[45] port <value> irq <irq>\"\n");
  printf("\t Use IRQ = 4 for 0x3F8|0x33h|0x3E8|0x2E8|0x220\n");
  printf("\t Use IRQ = 3 for 0x2F8|0x23h|0x2E|0x228\n");
  printf("\t\"setserial /dev/ttyS1[45] autoconfig\"\n");
  printf("\t\n");
  printf("\t%s\n", version);
  exit (1);
}

int main (int argc, char ** argv) {
  unsigned char regstring[5*NO_SMC_REGS+1], result;
  unsigned char smc_io_reg [NO_SMC_REGS];
  int fd,i, value, comm3, comm4;
  char *progname;
  unsigned char *myvalue;
  char intermediate[20];

  progname = argv[0];
  if (argc >1 ) print_usage(progname);

  fd = open("/proc/smcio",O_RDONLY|O_NONBLOCK, 0);
  if (fd < 0 ) {
    printf("Could not open /proc/smcio. Check that module is inserted\n");
    return -1;
  }
  /* We read some strang ascii string with termination! */
  result = read (fd, &regstring,5*NO_SMC_REGS+1);
  if (result != 5*NO_SMC_REGS+1) {
    printf("Didn't get valid entries from /proc/smcio\n");
    return -1;
  }
  /* Lets get the hex values back */
  
  /* Brute force, but I don't understand that better :-( */
  sscanf(regstring,"%i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i ", 
	 &smc_io_reg[0],
	 &smc_io_reg[1],
	 &smc_io_reg[2],
	 &smc_io_reg[3],
	 &smc_io_reg[4],
	 &smc_io_reg[5],
	 &smc_io_reg[6],
	 &smc_io_reg[7],
	 &smc_io_reg[8],
	 &smc_io_reg[9],
	 &smc_io_reg[10],
	 &smc_io_reg[11],
	 &smc_io_reg[12],
	 &smc_io_reg[13],
	 &smc_io_reg[14],
	 &smc_io_reg[15]);
  for (i=0; i < NO_SMC_REGS; i++) 
    printf("0x%.2x ",smc_io_reg[i]); 
  printf("\n");

  printf("Super-IO Chip is a FDC37C6%x\n",smc_io_reg[13]);
	 
  printf("IDE %s, %s mode\n",
	 (smc_io_reg[0] & 0x1) ? "enabled," : "disabled",
	 (smc_io_reg[0] & 0x2) ? "AT" : "XT");
  printf("Floppy %s and power %s \n", 
	 (smc_io_reg[0] & 0x10) ? "enabled" : "diabled",
	 (smc_io_reg[0] & 0x10) ? "on" : "off");
  printf("Oscillator Type %d\n", ((smc_io_reg[0] & 0x60)>>5));
  printf("Configuration %s\n", (smc_io_reg[0] & 0x80) ? "valid" : "invalid");
  
  switch (smc_io_reg[1] & 0x3) {
  case 1 : 
    value = 0x3bc;
    break;
  case 2 : 
    value = 0x378;
    break;
  case 3 : 
    value = 0x278;
    break;
  default:
    value = 0;
  }
  printf("Parallel Port power %s,", 
	 (smc_io_reg[1] & 0x4) ? "on" : "off");
  if (value == 0)
    printf ("disabled, ");
  else
    printf ("address 0x%3x,",value);
  printf("extended modes %sabled\n", (smc_io_reg[1] & 0x8) ? "en" : "dis");
  switch (smc_io_reg[4] & 0x3) {
  case 1 : 
    printf("\tEPP/SPP Mode");
    break;
  case 2 : 
    printf("\tECP Mode");
    break;
  case 3 : 
    printf("\tECP/EPP Mode");
    break;
  default:
    printf("\tStandard/SPP Mode");
  }
  printf(", EPP %s\n", (smc_io_reg[4] & 0x40) ? "1.9" : "1.7");
  printf("\tFifo Threshold %d\n", smc_io_reg[10]);
  

  printf("IRQ Polarity active %s\n",
	 (smc_io_reg[1] & 0x10)? "high, Push/Pull" : "low, Open Collector");
  printf("Configuration access %s\n", 
	 (smc_io_reg[1] & 0x80)? "allowed" : "disabled until reset");

  switch ((smc_io_reg[1] & 0x60) >> 5) {
  case 1 : 
    comm3 = 0x3e8;
    comm4 = 0x2e8;
    break;
  case 2 : 
    comm3 = 0x2e8;
    comm4 = 0x2e0;
    break;
  case 3 : 
    comm3 = 0x220;
    comm4 = 0x228;
    break;
  default:
    comm3 = 0x338;
    comm4 = 0x238;
  }
  switch (smc_io_reg[2] & 0x3) {
  case 1 : 
    value = 0x2f8;
    break;
  case 2 : 
    value = comm3;
    break;
  case 3 : 
    value = comm4;
    break;
  default:
    value = 0x3f8;
  }
  printf("UART1 power %s, %sabled at address 0x%3x, MIDI support %s\n", 
	 (smc_io_reg[2] & 0x8) ? "on" : "off",
	 (smc_io_reg[2] & 0x4) ? " en" : "dis", value,
	 (smc_io_reg[4] & 0x10) ? "on" : "off");
 
  switch ((smc_io_reg[2] & 0x30) >> 4) {
  case 1 : 
    value = 0x2f8;
    break;
  case 2 : 
    value = comm3;
    break;
  case 3 : 
    value = comm4;
    break;
  default:
    value = 0x3f8;
  }
  printf("UART2 power %s, %sabled at address 0x%3x, MIDI support %s\n", 
	 (smc_io_reg[2] & 0x80) ? "on" : "off",
	 (smc_io_reg[2] & 0x40) ? " en" : "dis", value,
	 (smc_io_reg[4] & 0x20) ? "on" : "off");

 
  return 0;
}



  
  
